#include "Wire.h"
#include "ZUNO_OLED_I2C.h"
//#include <ZUNO_OLED_FONT_NUMB16.h>
#include <ZUNO_OLED_FONT_NUMB24.h>
//#include <ZUNO_OLED_FONT_NUMB40.h>
// You can use another builtin fonts for numbers..._NUMB24, _NUMB40

#define SW_VERSION "SW 2.20"
#define DEVICE_NAME "Varmepump"

#ifndef OLED_SMALLFONT_MODIFIED
  // If this flag is not set, it means that the my modified ZUNO_OLED_I2C.cpp and .h have been overwritten
  // The flag is set in ZUNO_OLED_I2C.h
  // If flag missing then generate a compile error
void  Smallfontnotmodified();

#endif



// Set this flag to enable debug output in the serial monitor
#define DEBUGMODE

#ifdef DEBUGMODE
  #define DEBUG(s) Serial.print(s)
  #define DEBUGLN(s) Serial.println(s)
  #define DEBUG2(s,f) Serial.print(s,f)
  #define DEBUGLN2(s,f) Serial.println(s,f)

#else
  #define DEBUG(s)
  #define DEBUGLN(s)
  #define DEBUG2(s,f)
  #define DEBUGLN2(s,f)
#endif



#define PULSE_LENGTH 100


/* How many pulses between calculations
 * The higher the number, the smaller the variance
 * And how many seconds between calculations
 */
#define CALC_SPAN_PULSES 4
#define DISPLAY_INTERVAL 2UL


/* Energy meter ticks
 *
 * How many pulses/ticks per 1 kWh
 * Default setting: 1000
 */
#define TICKS_PER_KWH_PARAMETER 64
#define TICKS_PER_KWH_DEFAULT 1000UL
unsigned long ppkwh;

/* Power change reports
 *
 * Minimum change in consumed power that will result in sending new power report to the main controller.
 * 0 – reports are disabled
 * 1-100 (1-100%) – change in power
 * Default setting: 5 (5%)
 */
#define POWER_CHANGE_REPORT_PARAMETER 65
#define POWER_CHANGE_REPORT_DEFAULT 5UL
unsigned long p_changereport;

/* Time between power reports
 *
 * Minimum time that has to elapse before sending new power report to the main controller.
 * 0 – reports are disabled
 * 1-3600 (1-3600s) – report interval
 * Default setting: 30 (30s)
 */
#define POWER_TIME_REPORT_PARAMETER 66
#define POWER_TIME_REPORT_DEFAULT 30UL
unsigned long p_timereport;

/* Forced power reports
 *
 * Send power report even if no change to report.
 * 0 – reports are disabled
 * 1-3600 (1 – 3600s) – report interval
 * Default setting: 3600 (1 hour)
 */
#define POWER_TIME_REPORT_FORCED_PARAMETER 67
#define POWER_TIME_REPORT_FORCED_DEFAULT 600UL
unsigned long p_timereportforced;

/* Minimum power level
 *
 * If power level is below this level, then set it to zero
 * Default setting: 5
 */
#define POWER_MIN_VALUE_PARAMETER 68
#define POWER_MIN_VALUE_DEFAULT 5UL
unsigned long p_minvalue;

unsigned long currentpower;
unsigned long previouspower;
unsigned long powerincrease;
unsigned long powerdelta;
unsigned long previouspowerreport;


// Timestammps and counters
unsigned long pulsecount = 0;
unsigned long previouspulsecount = 0;
unsigned long pulsestart = 0;
unsigned long pulsespan = 0;
unsigned long previousdisplaytime = 0;
unsigned long currenttime = 0;
unsigned long timeoutpulse = 0;
unsigned long timeoutpulsespan = 0;
unsigned long pulsetime = 0;
unsigned long previouspulsetime = 0;


bool pulseactive = false;
bool sendpowerreport = true;

// Interrupt pin where the pulse counter is attached
int pulsepin = 3;

// Output LED which will blnk at same frequency as meter signal
s_pin signalpin = 14;

// Driver for the oled display
OLED oled;



#define ZUNO_CHANNEL_NUMBER_POWER 1


DWORD getter_power(){
  DEBUG("Getter - Power - Reported value = ");
  DEBUGLN(currentpower);
  return (DWORD)currentpower;
}


void resetter_power() {
  // This function must also reset the energy meter beaseuse the Fibaro HC2 does not call the resetter for the energy meter
  DEBUGLN("Resetter - Power");
  currentpower = 0;
  previouspower = 0;
  DEBUGLN("Calling resetter_energy");
 // resetter_energy(0);
}

/* Z-Wave channels
 *
 */
 //ZUNO_METER(TYPE, RESETABLE,  SCALE, SIZE, PRECISION, GETTER, RESETTER)
ZUNO_SETUP_CHANNELS(
  ZUNO_METER(
  ZUNO_METER_TYPE_ELECTRIC,
  METER_RESET_ENABLE,
  ZUNO_METER_ELECTRIC_SCALE_WATTS,
  METER_SIZE_FOUR_BYTES,
  METER_PRECISION_ZERO_DECIMALS,
  getter_power,
  resetter_power));



// Set up the interrupt handler on the ZEROX pin (3)
ZUNO_SETUP_ISR_ZEROX(pulse);


/* Interrupt proceure to be called when pulse is recevied
 *
 */
void pulse () {

  pulsecount++;

}


void setup() {


  #ifdef DEBUGMODE
  Serial.begin();
  #endif


  DEBUGLN("Starting");


  /* Set up values for counters
   * fetch from z-wave parameters
   */
  ppkwh = TICKS_PER_KWH_DEFAULT;
  p_changereport = POWER_CHANGE_REPORT_DEFAULT;
  p_timereport = POWER_TIME_REPORT_DEFAULT;
  p_timereportforced = POWER_TIME_REPORT_FORCED_DEFAULT;
  p_minvalue = POWER_MIN_VALUE_DEFAULT;

/*
  DEBUG("Z-wave parameters: ");
  DEBUG(ppkwh);
  DEBUG(p_changereport);
  DEBUG(", ");
  DEBUG(p_timereport);
  DEBUG(", ");
  DEBUG(p_timereportforced);
  DEBUG(", ");
  DEBUGLN(p_minvalue);
*/
  pulsecount = 0;

  // Set up output LED which will blnk at same frequency as meter signel
  pinMode(signalpin, OUTPUT);

  // Interrupt pin where the pulse counter is attached
  pinMode(pulsepin, INPUT);

  previouspowerreport = 0;

  currentpower = 0;
  previouspower = 0;
  pulsetime = 0;
  previouspulsetime = 0;
  timeoutpulse = 0;
  timeoutpulsespan = 0;
  sendpowerreport = true;


  // Initialize the oled display
  oled.begin();
  oled.clrscr();
  oled.setFont(SmallFont);
  oled.gotoXY(0,0);
  oled.setFont(SmallFont);
  oled.print("V");
  oled.write(228);
  oled.print("rmepump");
  //oled.print("Spis");

  oled.gotoXY(95,5);
  oled.print("Watt");
  oled.gotoXY(0,7);
  oled.print(SW_VERSION);

  // initiate the interrupt routine
  currenttime = millis();
  zunoExtIntMode(ZUNO_EXT_ZEROX, RISING);


}

void loop() {

  currenttime = millis();

  // Check if a pulse has been received
  if (pulsecount > previouspulsecount) {
    // Keep track of when the pulse started and switch on the LED
    pulsestart = currenttime;
    pulseactive = true;
    digitalWrite(signalpin,HIGH);

    DEBUG(SW_VERSION);
    DEBUG(" : ");
    DEBUGLN(pulsecount);
    DEBUG("currenttime= ");
    DEBUGLN(currenttime);
 /*   DEBUG(", pulsetime= ");
    DEBUG(pulsetime);
    DEBUG(", previouspulsetime =");
    DEBUG(previouspulsetime);
    DEBUG(", pulsespan = ");
    DEBUG(pulsespan);
    DEBUG(", pulse timeout span = ");
    DEBUG(timeoutpulsespan);
    DEBUG(", next timeout = ");
    DEBUGLN(timeoutpulse);

    DEBUG("Z-wave parameters: ");
    DEBUG(ppkwh);
    DEBUG(", ");
    DEBUG(p_changereport);
    DEBUG(", ");
    DEBUG(p_timereport);
    DEBUG(", ");
    DEBUG(p_timereportforced);
    DEBUG(", ");
    DEBUGLN(p_minvalue);
*/

    previouspulsecount = pulsecount;
    pulsespan = currenttime - previouspulsetime;
    currentpower = (3600*ppkwh)/pulsespan;

    // Adjust -1%
    currentpower = (currentpower * 99UL) / 100UL;

    previouspulsetime = currenttime;
    timeoutpulsespan = pulsespan + (pulsespan / 2);
    timeoutpulse = currenttime + timeoutpulsespan;

    DEBUG("currentpower= ");
    DEBUG(currentpower);
    DEBUG(" , previouspower= ");
    DEBUGLN(previouspower);

    if (currentpower > previouspower) {
      powerdelta = currentpower-previouspower;
    } else {
      powerdelta = previouspower-currentpower;
    }
    if (previouspower > 0) {
      powerincrease = 100*powerdelta / previouspower;
    } else {
      if (powerdelta > 0) {
        // Power has increased from zero, make sure we send a report
        powerincrease = p_changereport;
      } else {
        powerincrease = 0;
      }

    }

    // Check if power level has increased enough for a report
    if ((powerincrease >= p_changereport) && (p_changereport > 0)) {
      // Change in power level higher than report threshold. Send report
      DEBUG("Power change > ");
      DEBUG(p_changereport);
      DEBUGLN("%");
      sendpowerreport = true;

    }
    // If there is no change since previous report
    if (currentpower == previouspower){
      sendpowerreport = false;
    }
  }



  // Check if too long time has passed since the last pulse
  // This means that the power is lower than p_minvalue and should be interpreted as zero
  if ((currenttime > timeoutpulse) && (timeoutpulse > 0)){
      currentpower =   (3600*ppkwh)/(currenttime-previouspulsetime);
      timeoutpulse = timeoutpulse + timeoutpulsespan;

      if (currentpower != previouspower) {
        sendpowerreport = true;
      }

      DEBUG("Pulse timeout at ");
      DEBUG(currenttime);
      DEBUG(", previous pulse: ");
      DEBUG(previouspulsetime);
      DEBUG(", time since last pulse: ");
      DEBUG(currenttime-previouspulsetime);
      DEBUG(", power = ");
      DEBUG(currentpower);
      DEBUG(", previouspower = ");
      DEBUGLN(previouspower);

  }


  // Check if we should update the display
  if ((currenttime - previousdisplaytime) > (DISPLAY_INTERVAL*1000UL)) {

    // Show current power consumption on the OLED
    // Secure that the number is 5 positions long by padding space characters in front
    oled.setFont(zuno_font_numbers24);
    oled.gotoXY(0,3);


    if (currentpower < 10UL) {
      oled.print("     ");
    }
    else if (currentpower < 100UL) {
      oled.print("    ");
    }
    else if (currentpower < 1000UL) {
      oled.print("   ");
    }
    else if (currentpower < 10000UL) {
      oled.print("  ");
    }
    oled.print(currentpower);

    previousdisplaytime = currenttime;

  }


  // Check if enough time has passed since last report so we are allowed to send a report
  if ((currenttime - previouspowerreport) < (p_timereport*1000)) {
    // Too soon, don't send a report
    sendpowerreport = false;
  }


  // Check if we should force a report
  if (((currenttime - previouspowerreport) > (p_timereportforced*1000)) || (previouspowerreport == 0)){
    // Time to send a forced report
    DEBUG("Force send report. currenttime=");
    DEBUG(currenttime);
    DEBUG(", previousreport=");
    DEBUG(previouspowerreport);
    DEBUG(", delta=");
    DEBUG(currenttime - previouspowerreport);
    DEBUG(", limit=");
    DEBUGLN(p_timereportforced*1000);

    sendpowerreport = true;
  }


  if (sendpowerreport == true) {
    DEBUG("Sending power report. Power : ");
    DEBUGLN(currentpower);

    zunoSendReport(ZUNO_CHANNEL_NUMBER_POWER);
    previouspowerreport = currenttime;
    previouspower = currentpower;
    sendpowerreport = false;
  }


  // Check if the LED should be switched off
  if ((pulseactive == true) && ((currenttime - pulsestart) > PULSE_LENGTH)) {
    digitalWrite(signalpin, LOW);
    pulseactive = false;
  }

}
